/* -*- Mode: C; tab-width: 3; indent-tabs-mode: nil; c-basic-offset: 3 -*- */

/*
 * GImageView
 * Copyright (C) 2001-2002 Takuro Ashie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "pixbuf_loader.h"

#ifdef HAVE_GDK_PIXBUF

#include <gdk-pixbuf/gdk-pixbuf-loader.h>

#include "pixbuf_anim.h"
#include "fileutil.h"

void
cb_area_prepared (GdkPixbufLoader *loader,
                  gboolean *ret)
{
   g_return_if_fail (ret);

   *ret = TRUE;
}


void
cb_area_updated (GdkPixbufLoader *loader,
                 gint arg1,
                 gint arg2,
                 gint arg3,
                 gint arg4,
                 gboolean *ret)
{
   g_return_if_fail (ret);

   *ret = TRUE;
}


#ifndef USE_GTK2
void
cb_frame_done (GdkPixbufLoader *loader,
               gpointer arg1,
               gboolean *ret)
{
   g_return_if_fail (ret);

   *ret = TRUE;
}
#endif


GimvImage *
pixbuf_load (ImageLoader *loader, gpointer data)
{
   GimvIO *gio;
   GimvImage *image = NULL;
   GdkPixbufLoader *pixbuf_loader;
   guchar buf[512];
   gint buf_size = sizeof (buf) / sizeof (gchar);
   gint i;
   guint bytes;
   gboolean prepared = FALSE, updated = FALSE, frame_done = FALSE;
   gboolean cancel = FALSE;

   g_return_val_if_fail (loader, NULL);

   gio = image_loader_get_gio (loader);
   g_return_val_if_fail (gio, NULL);

   pixbuf_loader = gdk_pixbuf_loader_new ();
   g_return_val_if_fail (pixbuf_loader, NULL);

   /* set signals */
#ifdef USE_GTK2
   g_signal_connect (G_OBJECT (pixbuf_loader), "area-prepared",
                     G_CALLBACK (cb_area_prepared),
                     &prepared);
   g_signal_connect (G_OBJECT (pixbuf_loader), "area-updated",
                     G_CALLBACK (cb_area_updated),
                     &updated);
#else
   gtk_signal_connect (GTK_OBJECT (pixbuf_loader), "area-prepared",
                       GTK_SIGNAL_FUNC (cb_area_prepared),
                       &prepared);
   gtk_signal_connect (GTK_OBJECT (pixbuf_loader), "area-updated",
                       GTK_SIGNAL_FUNC (cb_area_updated),
                       &updated);
   gtk_signal_connect (GTK_OBJECT (pixbuf_loader), "frame-done",
                       GTK_SIGNAL_FUNC (cb_frame_done),
                       &frame_done);
#endif

   /* load */
   for (i = 0;; i++) {
      gimv_io_read (gio, buf, buf_size, &bytes);

      if ((gint) bytes > 0) {
#ifdef USE_GTK2
         gdk_pixbuf_loader_write (pixbuf_loader, buf, bytes, NULL);
#else
         gdk_pixbuf_loader_write (pixbuf_loader, buf, bytes);
#endif
      } else {
         break;
      }

      cancel = !image_loader_progress_update (loader);

      if (cancel) goto FUNC_END;
      if (!image_loader_load_as_animation(loader) && frame_done) break;
   }

   if (!prepared) goto FUNC_END;

#ifdef USE_GTK2
   if (image_loader_load_as_animation(loader)) {
#else
   if (image_loader_load_as_animation(loader) && frame_done) {
#endif
      GdkPixbufAnimation *anim;
      anim = gdk_pixbuf_loader_get_animation (pixbuf_loader);
      if (anim) {
         image = gimv_anim_new_from_gdk_pixbuf_animation (anim);
      }
   }

   if (!image) {
      image = gimv_image_new ();
      image->image = gdk_pixbuf_loader_get_pixbuf (pixbuf_loader);
      if (!image->image) {
         gimv_image_finalize (image);
         image = NULL;
      } else {
         gdk_pixbuf_ref (image->image);
      }
   }

 FUNC_END:
#ifdef USE_GTK2
   gdk_pixbuf_loader_close (pixbuf_loader, NULL);
   g_object_unref (G_OBJECT (pixbuf_loader));
#else
   gdk_pixbuf_loader_close (pixbuf_loader);
   gtk_object_unref (GTK_OBJECT (pixbuf_loader));
#endif

   return image;
}


GimvImage *
pixbuf_load_file (ImageLoader *loader, gpointer data)
{
   const gchar *filename;
   GimvImage *image = NULL;

   g_return_val_if_fail (loader, NULL);

   filename = image_loader_get_path (loader);
   g_return_val_if_fail (filename, NULL);

   if (!file_exists (filename))
      return NULL;

#ifdef USE_GTK2
   if (image_loader_load_as_animation(loader)) {
      GdkPixbufAnimation *anim;
      anim = gdk_pixbuf_animation_new_from_file (filename, NULL);
      if (anim) {
         image = gimv_anim_new_from_gdk_pixbuf_animation (anim);
         gdk_pixbuf_animation_unref (anim);
      }
   } else {
      image = gimv_image_new ();
      image->image = gdk_pixbuf_new_from_file (filename, NULL);
   }
#else
   if (image_loader_load_as_animation(loader)) {
      GdkPixbufAnimation *anim;
      anim = gdk_pixbuf_animation_new_from_file (filename);
      if (anim) {
         image = gimv_anim_new_from_gdk_pixbuf_animation (anim);
         gdk_pixbuf_animation_unref (anim);
      }
   } else {
      image = gimv_image_new ();
      image->image = gdk_pixbuf_new_from_file (filename);
   }
#endif /* USE_GTK2 */

   if (image && !image->image) {
      gimv_image_finalize (image);
      image = NULL;
   }

   return image;
}

#endif /* HAVE_GDK_PIXBUF */
