/* -*- Mode: C; tab-width: 3; indent-tabs-mode: nil; c-basic-offset: 3 -*- */

/*
 * GImageView
 * Copyright (C) 2001 Takuro Ashie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: gimv_bz_io.c,v 1.4.2.2 2002/11/22 10:38:29 makeinu Exp $
 */

#include "gimv_bz_io.h"

#ifdef HAVE_BZLIB

#include <stdio.h>
#include <string.h>
#include <bzlib.h>

#include "fileutil.h"

#ifndef BUF_SIZE
#  define BUF_SIZE 4096
#endif

#define BZIO_BLOCK_SIZE      65536   /* 64kB/1block */
#define BZIO_BLOCKS_STEP     64

static GimvIOStatus   gimv_bz_io_read  (GimvIO      *gio, 
                                        gchar       *buf, 
                                        guint        count,
                                        guint       *bytes_read);
static GimvIOStatus   gimv_bz_io_write (GimvIO      *gio, 
                                        const gchar *buf, 
                                        guint        count,
                                        guint       *bytes_written);
static GimvIOStatus   gimv_bz_io_seek  (GimvIO      *gio,
                                        glong        offset, 
                                        gint         whence);
static GimvIOStatus   gimv_bz_io_tell  (GimvIO      *gio,
                                        glong       *offset);
static void           gimv_bz_io_close (GimvIO      *gio);


typedef struct GimvBzIO_Tag GimvBzIO;

struct GimvBzIO_Tag
{
   GimvIO  parent;
   BZFILE *file;
   gchar **buflist;
   glong   pos;            /* virtual position */
   glong   bzpos;          /* stream position == read length*/
};

GimvIOFuncs gimv_bz_io_funcs =
{
   gimv_bz_io_read,
   gimv_bz_io_write,
   gimv_bz_io_seek,
   gimv_bz_io_tell,
   gimv_bz_io_close,
};


void
bz_io_read_from_stream_to_buf (GimvBzIO *bio)
{
   gint i, j;
   guint bytes = 1;

   for (i = 0; bytes > 0; i++) {
      if ((i != 0) && (i % BZIO_BLOCKS_STEP == 0)) {
         guint new_size;
         new_size = sizeof (gchar *) * BZIO_BLOCKS_STEP * (i / BZIO_BLOCKS_STEP + 1);
         bio->buflist = g_realloc (bio->buflist, new_size);
         for (j = 0; j < BZIO_BLOCKS_STEP; j++) {
            gint pos = i + j;
            bio->buflist[pos] = NULL;
         }
      }

      /* read stream */
      bio->buflist[i] = g_new (gchar, BZIO_BLOCK_SIZE);
      bytes = BZREAD (bio->file, bio->buflist[i], BZIO_BLOCK_SIZE);

      /* update parameter */
      bio->bzpos += bytes;

      if (bytes < BZIO_BLOCK_SIZE) break;
   }
}


GimvIO *
gimv_bz_io_new (const gchar *url, const gchar *mode)
{
   GimvBzIO *bio;
   GimvIO *gio;
   const gchar *filename = url;
   gint i;

   g_return_val_if_fail (url, NULL);

   if (!fileutil_extension_is (url, "bz2"))
      return NULL;

   bio = g_new0 (GimvBzIO, 1);
   gio  = (GimvIO *) bio;

   gimv_io_init (gio, url);
   gio->funcs = &gimv_bz_io_funcs;

   bio->file = BZOPEN (filename, (char *) mode);
   if (!bio->file) {
      gimv_io_close (gio);
      return NULL;
   }
   bio->buflist = g_new0 (gchar *, BZIO_BLOCKS_STEP);
   for (i = 0; i < BZIO_BLOCKS_STEP; i++)
      bio->buflist[i] = NULL;
   bio->pos   = 0;
   bio->bzpos = 0;

   bz_io_read_from_stream_to_buf (bio);

   return gio;
}


static GimvIOStatus
bz_io_read_from_buf (GimvBzIO *bio, 
                     gchar    *buf, 
                     guint     count,
                     guint    *bytes_read)
{
   gchar *dest;
   gint idx, offset;
   guint bytes;

   g_return_val_if_fail (bio->pos <= bio->bzpos, GIMV_IO_STATUS_ERROR);
   g_return_val_if_fail (bytes_read, GIMV_IO_STATUS_ERROR);

   *bytes_read = 0;

   idx = bio->pos / BZIO_BLOCK_SIZE;
   offset = bio->pos % BZIO_BLOCK_SIZE;
   bytes = MIN (count, BZIO_BLOCK_SIZE - offset);
   dest = buf;

   memcpy (dest, bio->buflist[idx] + offset, bytes);

   *bytes_read += bytes;
   bio->pos += bytes;
   count -= bytes;
   dest += bytes;

   while (count > 0) {
      idx++;

      bytes = MIN (count, BZIO_BLOCK_SIZE);

      memcpy (dest, bio->buflist[idx], bytes);

      *bytes_read += bytes;
      bio->pos += bytes;
      count -= bytes;
      dest += bytes;
   }

   return GIMV_IO_STATUS_NORMAL;
}


static GimvIOStatus
gimv_bz_io_read (GimvIO *gio, 
                 gchar  *buf, 
                 guint   count,
                 guint  *bytes_read)
{
   GimvBzIO *bio = (GimvBzIO *) gio;

   *bytes_read = 0;

   if (bio->pos == bio->bzpos) {
      return GIMV_IO_STATUS_NORMAL;
   } else if (bio->pos > bio->bzpos) {
      bio->pos = bio->bzpos;
      return GIMV_IO_STATUS_ERROR;
   }

   bz_io_read_from_buf (bio, buf, count, bytes_read);

   return GIMV_IO_STATUS_NORMAL;
}


static GimvIOStatus
gimv_bz_io_write (GimvIO      *gio, 
                  const gchar *buf, 
                  guint        count,
                  guint       *bytes_written)

{
   /* GimvBzIO *bio = (GimvBzIO *) gio; */

   /* FIXME */

   return GIMV_IO_STATUS_ERROR;
}


static GimvIOStatus
gimv_bz_io_seek  (GimvIO *gio,
                  glong   offset, 
                  gint    whence)
{
   GimvBzIO *bio = (GimvBzIO *) gio;
   glong pos;

   switch (whence) {
   case SEEK_CUR:
      pos = bio->pos + offset;
      break;
   case SEEK_SET:
      pos = offset;
      break;
   case SEEK_END:
      pos = bio->bzpos + offset;
      break;
   default:
      return GIMV_IO_STATUS_ERROR;
      break;
   }

   if (pos > bio->bzpos) {
      bio->pos = bio->bzpos;
      return GIMV_IO_STATUS_ERROR;
   } else if (pos < 0) {
      bio->pos = 0;
      return GIMV_IO_STATUS_ERROR;
   }

   bio->pos = pos;

   return GIMV_IO_STATUS_NORMAL;
}


static GimvIOStatus
gimv_bz_io_tell  (GimvIO *gio,
                  glong  *offset)
{
   GimvBzIO *bio = (GimvBzIO *) gio;

   g_return_val_if_fail (offset, GIMV_IO_STATUS_ERROR);

   *offset = bio->pos;

   return GIMV_IO_STATUS_NORMAL;
}


static void
gimv_bz_io_close  (GimvIO *gio)
{
   GimvBzIO *bio = (GimvBzIO *) gio;
   gint block_num, i;

   block_num = bio->bzpos / BZIO_BLOCK_SIZE + 1;
   for (i = 0; i < block_num; i++) {
      g_free (bio->buflist[i]);
   }
   g_free (bio->buflist);

   if (bio->file)
      BZCLOSE (bio->file);
}

#endif /* HAVE_BZLIB */
