/* -*- Mode: C; tab-width: 3; indent-tabs-mode: nil; c-basic-offset: 3 -*- */

/*
 * GImageView
 * Copyright (C) 2001 Takuro Ashie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: gimv_image.c,v 1.53.2.4 2003/01/28 20:46:45 makeinu Exp $
 */

#include "gimageview.h"

#include <stdio.h>
#include <string.h>

#include "gimv_image.h"
#include "gimv_anim.h"
#include "prefs.h"
#include "fileutil.h"
#include "gimv_io.h"
#include "image_saver.h"
#include "image_loader.h"

#if defined (HAVE_GDK_PIXBUF)
#  include <gdk-pixbuf/gdk-pixbuf.h>
#  include "pixbuf_utils.h"
typedef GdkPixbuf GimvRawImage;
#elif defined (HAVE_GDK_IMLIB)
#  include <gdk_imlib.h>
typedef GdkImlibImage GimvRawImage;
#endif /* HAVE_GDK_PIXBUF */



/****************************************************************************
 *
 *  File name extension list
 *
 ****************************************************************************/
/* for file name extension list */
typedef struct _img_ext
{
   gchar *ext;
   gchar *type;
} img_ext;


/* FIXME */
/* these values shuld be held by each plugins */
static img_ext img_exts[] =
{
   {"tif",  IMG_TIF},
   {"tiff", IMG_TIF},
   {"jpg",  IMG_JPG},
   {"jpe",  IMG_JPG},
   {"jpeg", IMG_JPG},
   {"png",  IMG_PNG},
   {"xbm",  IMG_XBM},
   {"xpm",  IMG_XPM},
   {"xwd",  IMG_XWD},
   {"ppm",  IMG_PPM},
   {"pgm",  IMG_PGM},
   {"pbm",  IMG_PBM},
   {"bmp",  IMG_BMP},
   {"pcx",  IMG_PCX},
   {"gif",  IMG_GIF},
   {"ico",  IMG_UNKNOWN},
   {"icon", IMG_UNKNOWN},
   {"mag",  IMG_MAG},
   {"tga",  IMG_TGA},
   {"xcf",  IMG_XCF},
#ifdef ENABLE_MNG
   {"mng",  IMG_MNG},
   {"jng",  IMG_JNG},
#endif
#ifdef ENABLE_SVG
   {"svg",  IMG_SVG},
#endif
#ifdef ENABLE_WMF
   {"wmf",  IMG_WMF},
#endif
   /*
     {"pic",  IMG_PICT},
     {"pct",  IMG_PICT},
     {"pict", IMG_PICT},
   */

   {NULL,   NULL}
};


/* FIXME: currentry, this array is referenced from image_info.c */
img_ext movie_exts[] =
{
#if ENABLE_XINE || ENABLE_MPLAYER
   {"mpg",  "MPEG"},
   {"mpeg", "MPEG"},
   {"m2v",  "MPEG"},
   {"vob",  "MPEG"},
   {"avi",  "AVI"},
   {"asf",  "ASF"},
   {"wmv",  "WMV"},
   {"qt",   "QuickTime"},
   {"mov",  "QuickTime"},
#endif /* ENABLE_XINE || ENABLE_MPLAYER */

   {NULL,   NULL}
};


static GList *img_system_ext_list_all = NULL;

static GList *img_system_ext_list = NULL;
static GHashTable *img_system_ext_table = NULL;

static GList *img_user_ext_list = NULL;
static GHashTable *img_user_ext_table = NULL;


static gboolean
sysdef_check_enable (gchar *ext, gchar **list)
{
   gint i;

   if (!list) return TRUE;

   for (i = 0; list[i]; i++) {
      if (list[i] && *list[i] && !strcmp (ext, list[i]))
         return FALSE;
   }

   return TRUE;
}


static void
free_image_types ()
{
   if (img_system_ext_list_all)
      g_list_free (img_system_ext_list_all);
   img_system_ext_list_all = NULL;

   if (img_system_ext_list)
      g_list_free (img_system_ext_list);
   img_system_ext_list = NULL;

   if (img_system_ext_table)
      g_hash_table_destroy (img_system_ext_table);
   img_system_ext_table = NULL;

   if (img_user_ext_table)
      g_hash_table_destroy (img_user_ext_table);
   img_user_ext_table = NULL;

   if (img_user_ext_list) {
      g_list_foreach (img_user_ext_list, (GFunc) g_free, NULL);
      g_list_free (img_user_ext_list);
   }
   img_user_ext_list = NULL;
}


void
gimv_image_update_image_types ()
{
   gchar **sysdef_disables = NULL;
   gchar **user_defs = NULL, **sections = NULL;
   gchar *ext, *type;
   gint i, j;
   img_ext *exts;
   img_ext *exts_array[] = {img_exts, movie_exts};

   free_image_types ();

   /* store system defined image types */

   if (conf.imgtype_disables && *conf.imgtype_disables) {
      sysdef_disables = g_strsplit (conf.imgtype_disables, ",", -1);
      if (sysdef_disables)
         for (i = 0; sysdef_disables[i]; i++)
            if (*sysdef_disables[i])
               g_strstrip (sysdef_disables[i]);
   }

   if (!img_system_ext_table)
      img_system_ext_table = g_hash_table_new (g_str_hash, g_str_equal);

   for (j = 0; j < sizeof (exts_array) / sizeof (img_ext *); j++) {
      exts = exts_array[j];
      for (i = 0; exts[i].ext; i++) {
         img_system_ext_list_all = g_list_append (img_system_ext_list_all,
                                                  exts[i].ext);

         if (sysdef_check_enable (exts[i].ext, sysdef_disables)) {
            img_system_ext_list = g_list_append (img_system_ext_list,
                                                 exts[i].ext);
         }

         g_hash_table_insert (img_system_ext_table,
                              exts[i].ext,
                              exts[i].type);
      }
   }

   if (sysdef_disables)
      g_strfreev (sysdef_disables);


   /* store user defined image types */

   if (conf.imgtype_user_defs && *conf.imgtype_user_defs)
      user_defs = g_strsplit (conf.imgtype_user_defs, ";", -1);
   if (!user_defs) return;

   if (!img_user_ext_table)
      img_user_ext_table = g_hash_table_new (g_str_hash, g_str_equal);

   for (i = 0; user_defs[i]; i++) {
      if (!*user_defs[i]) continue;

      sections = g_strsplit (user_defs[i], ",", 3);
      if (!sections) continue;
      g_strstrip (sections[0]);	 
      g_strstrip (sections[1]);	 
      g_strstrip (sections[2]);	 
      if (!*sections[0]) goto ERROR;
      /* if (filter_check_duplicate (sections[0], -1, FALSE)) goto ERROR; */

      if (!*sections[2] || strcasecmp (sections[2], "ENABLE")) goto ERROR;

      ext = g_strdup (sections[0]);
      if (!*sections[1])
         type = g_strdup ("UNKNOWN");
      else
         type = g_strdup (sections[1]);

      img_user_ext_list = g_list_append (img_user_ext_list, ext);
      g_hash_table_insert (img_user_ext_table, ext, type);

   ERROR:
      g_strfreev (sections);
   }

   g_strfreev (user_defs);
}



/****************************************************************************
 *
 *  GimvImage methods
 *
 ****************************************************************************/
void
gimv_image_init (void)
{
#if defined (HAVE_GDK_PIXBUF)
#elif defined (HAVE_GDK_IMLIB)
   /*
     GdkImlibInitParams *imlib_param;

     imlib_param = g_new0 (GdkImlibInitParams, 1);

     imlib_param->flags = PARAMS_REMAP | PARAMS_FASTRENDER;
     imlib_param->remap = 0;
     imlib_param->fastrender = 0;
     gdk_imlib_init_params(imlib_param);
     g_free (imlib_param);
   */
   gdk_imlib_init ();
   gtk_widget_push_visual (gdk_imlib_get_visual());
   gtk_widget_push_colormap (gdk_imlib_get_colormap());
#endif /* HAVE_GDK_PIXBUF */

   gimv_image_update_image_types ();
}


void
gimv_image_memset (GimvImage *image)
{
   image->image           = NULL;
   image->angle           = GIMV_IMAGE_ROTATE_0;
   image->flags           = 0;
   image->ref_count       = 1;
   image->comments        = NULL;
   image->additional_data = NULL;
}


GimvImage *
gimv_image_new (void)
{
   GimvImage *image = g_new0 (GimvImage, 1);

   gimv_image_memset (image);

   return image;
}


GimvImage *
gimv_image_load_file (const gchar *filename, gboolean animation)
{
   ImageLoader *loader;
   GimvImage *image;

   loader = image_loader_new_with_file_name (filename);
   if (!loader) return NULL;

   image_loader_set_as_animation (loader, animation);
   image_loader_load (loader);
   image = image_loader_get_image (loader);
   if (image)
      gimv_image_ref (image);
   image_loader_unref (loader);

   return image;
}


#if defined (HAVE_GDK_PIXBUF)
static void
free_rgb_buffer (guchar *pixels, gpointer data)
{
   g_free(pixels);
}
#endif /* HAVE_GDK_PIXBUF */


GimvImage *
gimv_image_create_from_data (guchar *data, gint width, gint height, gboolean alpha)
{
   GimvImage *image;

   image = gimv_image_new ();

#if defined (HAVE_GDK_PIXBUF)
   {
      gint bytes = 3;

      if (alpha)
         bytes = 4;

      image->image = gdk_pixbuf_new_from_data (data,
                                               GDK_COLORSPACE_RGB, alpha, 8,
                                               width, height, bytes * width,
                                               free_rgb_buffer, NULL);
   }
#elif defined (HAVE_GDK_IMLIB)
   image->image = gdk_imlib_create_image_from_data (data, NULL, width, height);
   g_free (data);
#endif /* HAVE_GDK_PIXBUF */

   if (!image->image) {
      gimv_image_finalize (image);
      image = NULL;
   }

   return image;
}


GimvImage *
gimv_image_create_from_drawable (GdkDrawable *drawable, gint x, gint y,
                                 gint width, gint height)
{
   GimvImage *image;

   image = gimv_image_new ();

#if defined (HAVE_GDK_PIXBUF)
   {
      GdkColormap *cmap = gdk_colormap_get_system ();
      image->image = gdk_pixbuf_get_from_drawable (NULL, drawable, cmap,
                                                   x , y, 0, 0,
                                                   width, height);
   }
#elif defined (HAVE_GDK_IMLIB)
   image->image = gdk_imlib_create_image_from_drawable (drawable, NULL,
                                                        x, y,
                                                        width, height);
#endif /* HAVE_GDK_PIXBUF */

   if (!image->image) {
      gimv_image_finalize (image);
      image = NULL;
   }

   return image;
}


void
gimv_image_add_comment (GimvImage *image,
                        const gchar *key,
                        const gchar *val)
{
   gchar *orig_key, *old_value;
   gboolean success;

   g_return_if_fail (image);
   g_return_if_fail (key && *key);
   g_return_if_fail (val);

   if (!image->comments)
      image->comments = g_hash_table_new (g_str_hash, g_str_equal);

   success = g_hash_table_lookup_extended (image->comments, key,
                                           (gpointer) &orig_key,
                                           (gpointer) &old_value);
   if (success) {
      g_hash_table_remove (image->comments, key);
      g_free (orig_key);
      g_free (old_value);
   }

   g_hash_table_insert (image->comments, g_strdup (key), g_strdup (val));
}


GimvImage *
gimv_image_rotate_90 (GimvImage *image,
                      gboolean counter_clockwise)
{
   GimvRawImage *src_rawimage, *dest_rawimage;

   g_return_val_if_fail (image, NULL);
   g_return_val_if_fail (image->image, NULL);

   src_rawimage = (GimvRawImage *) image->image;

#if defined (HAVE_GDK_PIXBUF)
   dest_rawimage = pixbuf_copy_rotate_90 (src_rawimage, counter_clockwise);
   gdk_pixbuf_unref (src_rawimage);
#elif defined (HAVE_GDK_IMLIB)
   if (counter_clockwise) {
      gdk_imlib_flip_image_horizontal (src_rawimage);
      gdk_imlib_rotate_image (src_rawimage, 1);
   } else {
      gdk_imlib_flip_image_vertical (src_rawimage);
      gdk_imlib_rotate_image (src_rawimage, 1);
   }
   dest_rawimage = src_rawimage;
#endif /* HAVE_GDK_PIXBUF */

   image->image = dest_rawimage;

   if (counter_clockwise)
      image->angle += GIMV_IMAGE_ROTATE_90;
   else
      image->angle += GIMV_IMAGE_ROTATE_270;

   image->angle %= 4;

   if (image->angle < 0)
      image->angle += 4;

   return image;
}


GimvImage *
gimv_image_rotate_180 (GimvImage *image)
{
   GimvRawImage *src_rawimage, *dest_rawimage;

   g_return_val_if_fail (image, NULL);
   g_return_val_if_fail (image->image, NULL);

   src_rawimage = (GimvRawImage *) image->image;

#if defined (HAVE_GDK_PIXBUF)
   dest_rawimage = pixbuf_copy_mirror (src_rawimage, TRUE, TRUE);
   gdk_pixbuf_unref (src_rawimage);
#elif defined (HAVE_GDK_IMLIB)
   gdk_imlib_flip_image_vertical (src_rawimage);
   gdk_imlib_flip_image_horizontal (src_rawimage);
   dest_rawimage = src_rawimage;
#endif /* HAVE_GDK_PIXBUF */

   image->image = dest_rawimage;

   image->angle += GIMV_IMAGE_ROTATE_180;
   image->angle %= 4;
   if (image->angle < 0)
      image->angle += 4;

   return image;
}


GimvImage *
gimv_image_rotate (GimvImage *image, GimvImageAngle abs_angle)
{
   gint rotate, angle;

   g_return_val_if_fail (image, NULL);

   if (image->angle < 0 || image->angle > 3) {
      image->angle %= 4;
      if (image->angle < 0)
         image->angle += 4;
   }

   if (abs_angle < 0 || abs_angle > 3) {
      angle = abs_angle % 4;
      if (angle < 0)
         angle += 4;
   } else {
      angle = abs_angle;
   }

   if (image->angle == angle)
      return image;

   rotate = angle - image->angle;
   if (rotate < 0)
      rotate += 4;

   switch (rotate) {
   case GIMV_IMAGE_ROTATE_90:
      gimv_image_rotate_90 (image, TRUE);
      break;
   case GIMV_IMAGE_ROTATE_180:
      gimv_image_rotate_180 (image);
      break;
   case GIMV_IMAGE_ROTATE_270:
      gimv_image_rotate_90 (image, FALSE);
      break;
   default:
      break;
   }

   return image;
}


void
gimv_image_get_pixmap_and_mask (GimvImage  *image,
                                GdkPixmap **pixmap_return,
                                GdkBitmap **mask_return)
{
   GimvRawImage *rawimage;

   g_return_if_fail (image);
   g_return_if_fail (image->image);
   g_return_if_fail (pixmap_return);
   g_return_if_fail (mask_return);

   rawimage = (GimvRawImage *) image->image;

#if defined (HAVE_GDK_PIXBUF)
   gdk_pixbuf_render_pixmap_and_mask (rawimage, pixmap_return, mask_return, 64);
#elif defined (HAVE_GDK_IMLIB)
   *pixmap_return = gdk_imlib_move_image (rawimage);
   *mask_return = gdk_imlib_move_mask (rawimage);
#endif /* HAVE_GDK_PIXBUF */
}


void
gimv_image_free_pixmap_and_mask (GdkPixmap *pixmap,
                                 GdkBitmap *mask)
{
#if defined (HAVE_GDK_PIXBUF)
   if (pixmap) gdk_pixmap_unref (pixmap);
   if (mask) gdk_bitmap_unref (mask);
#elif defined (HAVE_GDK_IMLIB)
   gdk_imlib_free_pixmap (pixmap);
   gdk_imlib_free_bitmap (mask);
#endif /* HAVE_GDK_PIXBUF */
}


gboolean
gimv_image_is_scalable (GimvImage *image)
{
   g_return_val_if_fail (image, FALSE);

   return image->flags & GIMV_IMAGE_VECTOR_FLAGS;
}


GimvImage *
gimv_image_scale_and_dump (GimvImage *image,
                           gint width, gint height)
{
   GimvImage *dest_image;
   GimvRawImage *src_rawimage, *dest_rawimage;

   g_return_val_if_fail (image, NULL);
   g_return_val_if_fail (image->image, NULL);

   src_rawimage = (GimvRawImage *) image->image;

   /*
     if (gimv_image_is_scalable (image)) {
     } else {
     }
   */

#if defined (HAVE_GDK_PIXBUF)
   dest_rawimage = gdk_pixbuf_scale_simple (src_rawimage, width, height,
                                            conf.interpolation);
#elif defined (HAVE_GDK_IMLIB)
   dest_rawimage = gdk_imlib_clone_scaled_image (src_rawimage, width, height);
#endif /* HAVE_GDK_PIXBUF */

   dest_image = gimv_image_new ();
   dest_image->image = dest_rawimage;

   return dest_image;
}


void
gimv_image_scale_get_pixmap (GimvImage *image,
                             gint width, gint height,
                             GdkPixmap **pixmap_return,
                             GdkBitmap **mask_return)
{
   g_return_if_fail (image);
   g_return_if_fail (image->image);
   g_return_if_fail (pixmap_return);
   g_return_if_fail (mask_return);

#if defined (HAVE_GDK_PIXBUF)
   {
      GimvImage *dest_image;
      dest_image = gimv_image_scale_and_dump (image, width, height);
      gimv_image_get_pixmap_and_mask (dest_image, pixmap_return, mask_return);
      gimv_image_finalize (dest_image);
   }
#elif defined (HAVE_GDK_IMLIB)
   gdk_imlib_render (image->image, width, height);
   gimv_image_get_pixmap_and_mask (image, pixmap_return, mask_return);
#endif
}


void
gimv_image_get_size (GimvImage *image, gint *width, gint *height)
{
   GimvRawImage *rawimage;

   g_return_if_fail (width && height);
   *width = *height = -1;

   g_return_if_fail (image);
   g_return_if_fail (image->image);

   rawimage = (GimvRawImage *) image->image;

#if defined (HAVE_GDK_PIXBUF)
   *width  = gdk_pixbuf_get_width  (rawimage);
   *height = gdk_pixbuf_get_height (rawimage);
#elif defined (HAVE_GDK_IMLIB)
   *width  = rawimage->rgb_width;
   *height = rawimage->rgb_height;
#endif /* HAVE_GDK_PIXBUF */
}


gint
gimv_image_width (GimvImage *image)
{
   GimvRawImage *rawimage;

   g_return_val_if_fail (image, -1);
   g_return_val_if_fail (image->image, -1);

   rawimage = (GimvRawImage *) image->image;

#if defined (HAVE_GDK_PIXBUF)
   return gdk_pixbuf_get_width  (rawimage);
#elif defined (HAVE_GDK_IMLIB)
   return rawimage->rgb_width;
#endif /* HAVE_GDK_PIXBUF */
}


gint
gimv_image_height (GimvImage *image)
{
   GimvRawImage *rawimage;

   g_return_val_if_fail (image, -1);
   g_return_val_if_fail (image->image, -1);

   rawimage = (GimvRawImage *) image->image;

#if defined (HAVE_GDK_PIXBUF)
   return gdk_pixbuf_get_height (rawimage);
#elif defined (HAVE_GDK_IMLIB)
   return rawimage->rgb_height;
#endif /* HAVE_GDK_PIXBUF */
}


gint
gimv_image_depth (GimvImage *image)
{
   GimvRawImage *rawimage;

   g_return_val_if_fail (image, -1);
   g_return_val_if_fail (image->image, -1);

   rawimage = (GimvRawImage *) image->image;

#if defined (HAVE_GDK_PIXBUF)
   return gdk_pixbuf_get_bits_per_sample (rawimage);
#elif defined (HAVE_GDK_IMLIB)
   return 8;
#endif /* HAVE_GDK_PIXBUF */
}


gboolean
gimv_image_has_alpha (GimvImage *image)
{
   GimvRawImage *rawimage;

   g_return_val_if_fail (image, FALSE);
   g_return_val_if_fail (image->image, FALSE);

   rawimage = (GimvRawImage *) image->image;

#if defined (HAVE_GDK_PIXBUF)
   return gdk_pixbuf_get_has_alpha (rawimage);
#elif defined (HAVE_GDK_IMLIB)
   return FALSE;
#endif /* HAVE_GDK_PIXBUF */
}


gboolean
gimv_image_can_alpha (GimvImage *image)
{
#if defined (HAVE_GDK_PIXBUF)
   return TRUE;
#elif defined (HAVE_GDK_IMLIB)
   return FALSE;
#endif /* HAVE_GDK_PIXBUF */
}


gint
gimv_image_rowstride (GimvImage *image)
{
   GimvRawImage *rawimage;

   g_return_val_if_fail (image, -1);
   g_return_val_if_fail (image->image, -1);

   rawimage = (GimvRawImage *) image->image;

#if defined (HAVE_GDK_PIXBUF)
   return gdk_pixbuf_get_rowstride (rawimage);
#elif defined (HAVE_GDK_IMLIB)
   if (!rawimage)
      return 0;
   else
      return rawimage->rgb_width * 3;
#endif /* HAVE_GDK_PIXBUF */
}


guchar *
gimv_image_get_pixels (GimvImage *image)
{
   GimvRawImage *rawimage;

   g_return_val_if_fail (image, NULL);
   g_return_val_if_fail (image->image, NULL);

   rawimage = (GimvRawImage *) image->image;

#if defined (HAVE_GDK_PIXBUF)
   return gdk_pixbuf_get_pixels (rawimage);
#elif defined (HAVE_GDK_IMLIB)
   if (!rawimage)
      return NULL;
   else
      return rawimage->rgb_data;
#endif /* HAVE_GDK_PIXBUF */
}


const gchar *
gimv_image_get_comment (GimvImage *image, const gchar *key)
{
   g_return_val_if_fail (image, NULL);

   if (!image->comments) return NULL;

   return g_hash_table_lookup (image->comments, key);
}


GimvImage  *
gimv_image_ref (GimvImage *image)
{
   g_return_val_if_fail (image, NULL);

   image->ref_count++;

   return image;
}


void
gimv_image_unref (GimvImage *image)
{
   g_return_if_fail (image);
   g_return_if_fail (image->ref_count > 0);

   image->ref_count--;

   if (image->ref_count < 1) {
      gimv_image_finalize (image);
   }
}


static void
free_comment (gpointer key, gpointer value, gpointer data)
{
   g_free (key);
   g_free (value);
}


void
gimv_image_free_comments (GimvImage *image)
{
   g_return_if_fail (image);

   if (!image->comments) return;

   g_hash_table_foreach (image->comments,
                         (GHFunc) free_comment,
                         image);
   g_hash_table_destroy (image->comments);
   image->comments = NULL;
}


void
gimv_image_finalize (GimvImage *image)
{
   GimvRawImage *rawimage;

   g_return_if_fail (image);

   rawimage = (GimvRawImage *) image->image;

   if (gimv_image_is_anim (image))
      gimv_anim_delete_private_data ((GimvAnim *) image);

#if defined (HAVE_GDK_PIXBUF)
   if (rawimage)
      gdk_pixbuf_unref (rawimage);
#elif defined (HAVE_GDK_IMLIB)
   gdk_imlib_kill_image (rawimage);
#endif /* HAVE_GDK_PIXBUF */

   gimv_image_free_comments (image);

   g_free (image);
}


gboolean
gimv_image_save_file  (GimvImage   *image,
                       const gchar *filename,
                       const gchar *format,
                       gpointer     data,
                       gboolean     use_specific_data)
{
   return image_saver_save_image (image, filename, format,
                                  data, use_specific_data);
}


/*
 *  gimv_image_detect_type_by_ext:
 *     @ Detect image format by filename extension.
 *
 *  str    : Image file name.
 *  Return : Image type by string.
 */
gchar *
gimv_image_detect_type_by_ext (const gchar *str)
{
   gchar *ext = fileutil_get_extention (str);
   GList *list;

   if (!ext) return NULL;

   for (list = g_list_first (img_system_ext_list); list;
        list = g_list_next (list))
   {
      gchar *sysext = list->data;

      if (!sysext) continue;

      if (!g_strcasecmp(ext, sysext)) {
         gchar *type = g_hash_table_lookup (img_system_ext_table, sysext);
         return type;
      }
   }

   for (list = g_list_first (img_user_ext_list); list;
        list = g_list_next (list))
   {
      gchar *userext = list->data;

      if (!userext) continue;

      if (!g_strcasecmp(ext, userext)) {
         gchar *type = g_hash_table_lookup (img_user_ext_table, userext);
         return type;
      }
   }

   return NULL;
}


GList *
gimv_image_get_sysdef_ext_list_all ()
{
   return img_system_ext_list_all;
}


GList *
gimv_image_get_sysdef_ext_list ()
{
   return img_system_ext_list;
}


GList *
gimv_image_get_userdef_ext_list ()
{
   return img_user_ext_list;
}


gchar *
gimv_image_get_type_from_ext (gchar *ext)
{
   gchar *type;

   g_return_val_if_fail (ext && *ext, NULL);

   type = g_hash_table_lookup (img_system_ext_table, ext);

   if (!type)
      type = g_hash_table_lookup (img_user_ext_table, ext);

   return type;
}


gboolean
gimv_image_is_anim (GimvImage *image)
{
   g_return_val_if_fail (image, FALSE);

   if (image->flags & GIMV_IMAGE_ANIMATION_FLAG)
      return TRUE;

   return FALSE;
}



/****************************************************************************
 *  FIXME!!
 ****************************************************************************/
static void
rgb_to_hsv (gint *r,
            gint *g,
            gint *b)
{
   gint red, green, blue;
   gfloat h, s, v;
   gint min, max;
   gint delta;

   h = 0.0;

   red = *r;
   green = *g;
   blue = *b;

   if (red > green) {
      if (red > blue)
         max = red;
      else
         max = blue;

      if (green < blue)
         min = green;
      else
         min = blue;
   } else {
      if (green > blue)
         max = green;
      else
         max = blue;

      if (red < blue)
         min = red;
      else
         min = blue;
   }

   v = max;

   if (max != 0)
      s = ((max - min) * 255) / (gfloat) max;
   else
      s = 0;

   if (s == 0) {
      h = 0;
   } else {
      delta = max - min;
      if (red == max)
         h = (green - blue) / (gfloat) delta;
      else if (green == max)
         h = 2 + (blue - red) / (gfloat) delta;
      else if (blue == max)
         h = 4 + (red - green) / (gfloat) delta;
      h *= 42.5;

      if (h < 0)
         h += 255;
      if (h > 255)
         h -= 255;
   }

   *r = h;
   *g = s;
   *b = v;
}


static void
hsv_to_rgb (gint *h,
            gint *s,
            gint *v)
{
   gfloat hue, saturation, value;
   gfloat f, p, q, t;

   if (*s == 0) {
      *h = *v;
      *s = *v;
      *v = *v;
   } else {
      hue = *h * 6.0 / 255.0;
      saturation = *s / 255.0;
      value = *v / 255.0;

      f = hue - (gint) hue;
      p = value * (1.0 - saturation);
      q = value * (1.0 - (saturation * f));
      t = value * (1.0 - (saturation * (1.0 - f)));

      switch ((gint) hue) {
      case 0:
         *h = value * 255;
         *s = t * 255;
         *v = p * 255;
         break;
      case 1:
         *h = q * 255;
         *s = value * 255;
         *v = p * 255;
         break;
      case 2:
         *h = p * 255;
         *s = value * 255;
         *v = t * 255;
         break;
      case 3:
         *h = p * 255;
         *s = q * 255;
         *v = value * 255;
         break;
      case 4:
         *h = t * 255;
         *s = p * 255;
         *v = value * 255;
         break;
      case 5:
         *h = value * 255;
         *s = p * 255;
         *v = q * 255;
         break;
      }
   }
}


static void
rgb_to_hls (gint *r,
            gint *g,
            gint *b)
{
   gint red, green, blue;
   gfloat h, l, s;
   gint min, max;
   gint delta;

   red = *r;
   green = *g;
   blue = *b;

   if (red > green) {
      if (red > blue)
         max = red;
      else
         max = blue;

      if (green < blue)
         min = green;
      else
         min = blue;
   } else {
      if (green > blue)
         max = green;
      else
         max = blue;

      if (red < blue)
         min = red;
      else
         min = blue;
   }

   l = (max + min) / 2.0;

   if (max == min) {
      s = 0.0;
      h = 0.0;
   } else {
      delta = (max - min);

      if (l < 128)
         s = 255 * (gfloat) delta / (gfloat) (max + min);
      else
         s = 255 * (gfloat) delta / (gfloat) (511 - max - min);

      if (red == max)
         h = (green - blue) / (gfloat) delta;
      else if (green == max)
         h = 2 + (blue - red) / (gfloat) delta;
      else
         h = 4 + (red - green) / (gfloat) delta;

      h = h * 42.5;

      if (h < 0)
         h += 255;
      if (h > 255)
         h -= 255;
   }

   *r = h;
   *g = l;
   *b = s;
}

static gint
hls_value (gfloat n1,
           gfloat n2,
           gfloat hue)
{
   gfloat value;

   if (hue > 255)
      hue -= 255;
   else if (hue < 0)
      hue += 255;
   if (hue < 42.5)
      value = n1 + (n2 - n1) * (hue / 42.5);
   else if (hue < 127.5)
      value = n2;
   else if (hue < 170)
      value = n1 + (n2 - n1) * ((170 - hue) / 42.5);
   else
      value = n1;

   return (gint) (value * 255);
}


static void
hls_to_rgb (gint *h,
            gint *l,
            gint *s)
{
   gfloat hue, lightness, saturation;
   gfloat m1, m2;

   hue = *h;
   lightness = *l;
   saturation = *s;

   if (saturation == 0) {
      /*  achromatic case  */
      *h = lightness;
      *l = lightness;
      *s = lightness;
   } else {
      if (lightness < 128)
         m2 = (lightness * (255 + saturation)) / 65025.0;
      else
         m2 = (lightness + saturation - (lightness * saturation)/255.0) / 255.0;

      m1 = (lightness / 127.5) - m2;

      /*  chromatic case  */
      *h = hls_value (m1, m2, hue + 85);
      *l = hls_value (m1, m2, hue);
      *s = hls_value (m1, m2, hue - 85);
   }
}


static void
pixel_apply_layer_mode(guchar *src1,
                       guchar *src2,
                       guchar *dest,
                       gint b1,
                       gint b2,
                       gboolean ha1,
                       gboolean ha2,
                       GimvImageLayerMode mode)
{
   static gint b, alpha, t, t2, red1, green1, blue1, red2, green2, blue2;
	
   alpha = (ha1 || ha2) ? MAX(b1, b2) - 1 : b1;
   switch (mode) {
   case DISSOLVE_MODE:
      for (b = 0; b < alpha; b++)
         dest[b] = src2[b];
      /*  dissolve if random value is > opacity  */
      t = (rand() & 0xFF);
      dest[alpha] = (t > src2[alpha]) ? 0 : src2[alpha];
      break;

   case MULTIPLY_MODE:
      for (b = 0; b < alpha; b++)
         dest[b] = (src1[b] * src2[b]) / 255;
      if (ha1 && ha2)
         dest[alpha] = MIN(src1[alpha], src2[alpha]);
      else if (ha2)
         dest[alpha] = src2[alpha];
      break;

   case SCREEN_MODE:
      for (b = 0; b < alpha; b++)
         dest[b] = 255 - ((255 - src1[b]) * (255 - src2[b])) / 255;
      if (ha1 && ha2)
         dest[alpha] = MIN(src1[alpha], src2[alpha]);
      else if (ha2)
         dest[alpha] = src2[alpha];
      break;

   case OVERLAY_MODE:
      for (b = 0; b < alpha; b++) {
         /* screen */
         t = 255 - ((255 - src1[b]) * (255 - src2[b])) / 255;
         /* mult */
         t2 = (src1[b] * src2[b]) / 255;
         dest[b] = (t * src1[b] + t2 * (255 - src1[b])) / 255;
      }
      if (ha1 && ha2)
         dest[alpha] = MIN(src1[alpha], src2[alpha]);
      else if (ha2)
         dest[alpha] = src2[alpha];
      break;

   case DIFFERENCE_MODE:
      for (b = 0; b < alpha; b++) {
         t = src1[b] - src2[b];
         dest[b] = (t < 0) ? -t : t;
      }
      if (ha1 && ha2)
         dest[alpha] = MIN(src1[alpha], src2[alpha]);
      else if (ha2)
         dest[alpha] = src2[alpha];
      break;

   case ADDITION_MODE:
      for (b = 0; b < alpha; b++) {
         t = src1[b] + src2[b];
         dest[b] = (t > 255) ? 255 : t;
      }
      if (ha1 && ha2)
         dest[alpha] = MIN(src1[alpha], src2[alpha]);
      else if (ha2)
         dest[alpha] = src2[alpha];
      break;

   case SUBTRACT_MODE:
      for (b = 0; b < alpha; b++) {
         t = (gint)src1[b] - (gint)src2[b];
         dest[b] = (t < 0) ? 0 : t;
      }
      if (ha1 && ha2)
         dest[alpha] = MIN(src1[alpha], src2[alpha]);
      else if (ha2)
         dest[alpha] = src2[alpha];
      break;

   case DARKEN_ONLY_MODE:
      for (b = 0; b < alpha; b++) {
         t = src1[b];
         t2 = src2[b];
         dest[b] = (t < t2) ? t : t2;
      }
      if (ha1 && ha2)
         dest[alpha] = MIN(src1[alpha], src2[alpha]);
      else if (ha2)
         dest[alpha] = src2[alpha];
      break;

   case LIGHTEN_ONLY_MODE:
      for (b = 0; b < alpha; b++) {
         t = src1[b];
         t2 = src2[b];
         dest[b] = (t < t2) ? t2 : t;
      }
      if (ha1 && ha2)
         dest[alpha] = MIN(src1[alpha], src2[alpha]);
      else if (ha2)
         dest[alpha] = src2[alpha];
      break;

   case HUE_MODE:
   case SATURATION_MODE:
   case VALUE_MODE:
      red1 = src1[0]; green1 = src1[1]; blue1 = src1[2];
      red2 = src2[0]; green2 = src2[1]; blue2 = src2[2];
      rgb_to_hsv (&red1, &green1, &blue1);
      rgb_to_hsv (&red2, &green2, &blue2);
      switch (mode) {
      case HUE_MODE:
         red1 = red2;
         break;
      case SATURATION_MODE:
         green1 = green2;
         break;
      case VALUE_MODE:
         blue1 = blue2;
         break;
      default:
         break;
      }
      hsv_to_rgb (&red1, &green1, &blue1);
      dest[0] = red1; dest[1] = green1; dest[2] = blue1;
      if (ha1 && ha2)
         dest[3] = MIN(src1[3], src2[3]);
      else if (ha2)
         dest[3] = src2[3];
      break;

   case COLOR_MODE:
      red1 = src1[0]; green1 = src1[1]; blue1 = src1[2];
      red2 = src2[0]; green2 = src2[1]; blue2 = src2[2];
      rgb_to_hls (&red1, &green1, &blue1);
      rgb_to_hls (&red2, &green2, &blue2);
      red1 = red2;
      blue1 = blue2;
      hls_to_rgb (&red1, &green1, &blue1);
      dest[0] = red1; dest[1] = green1; dest[2] = blue1;
      if (ha1 && ha2)
         dest[3] = MIN(src1[3], src2[3]);
      else if (ha2)
         dest[3] = src2[3];
      break;

   case NORMAL_MODE:
   default:
      for (b = 0; b < b2; b++) dest[b] = src2[b];
      break;
   }
}

static gint bgcolor_red   = 255;
static gint bgcolor_green = 255;
static gint bgcolor_blue  = 255;

gboolean
gimv_image_add_layer (guchar *buffer,
                      gint width,
                      gint left,
                      gint components,
                      gint pass,
                      GimvImageLayerMode mode,
                      guchar *rgbbuffer)
{
   gint i, j, k;
   static gint alpha, cur_red, cur_green, cur_blue;
   static guchar src1[4], src2[4], dest[4];

   if (pass <= 0) {
      cur_red   = bgcolor_red;
      cur_green = bgcolor_green;
      cur_blue  = bgcolor_blue;
   }

   for (i = 0, j = 0, k = 0; i < width; i++) {
      if (pass > 0) {
         cur_red   = rgbbuffer[k];
         cur_green = rgbbuffer[k + 1];
         cur_blue  = rgbbuffer[k + 2];
      }

      switch (components) {
      case 1:
         rgbbuffer[k++] = buffer[j];
         rgbbuffer[k++] = buffer[j];
         rgbbuffer[k++] = buffer[j++];
         break;

      case 2:
         if (pass > 0 && mode > 0) {
            src1[0] = cur_red;
            src1[1] = cur_green;
            src1[2] = cur_blue;
            src1[3] = 255;
            src2[0] = buffer[j];
            src2[1] = buffer[j];
            src2[2] = buffer[j];
            src2[3] = buffer[j + 1];
            pixel_apply_layer_mode(src1, src2, dest, 4, 4, TRUE, TRUE, mode);
            alpha = dest[3];
            rgbbuffer[k++] = dest[0] * alpha / 255 + cur_red   * (255 - alpha) / 255;
            rgbbuffer[k++] = dest[1] * alpha / 255 + cur_green * (255 - alpha) / 255;
            rgbbuffer[k++] = dest[2] * alpha / 255 + cur_blue  * (255 - alpha) / 255;
         } else {
            alpha = buffer[j + 1];
            rgbbuffer[k++] = buffer[j] * alpha / 255 + cur_red   * (255 - alpha) / 255;
            rgbbuffer[k++] = buffer[j] * alpha / 255 + cur_green * (255 - alpha) / 255;
            rgbbuffer[k++] = buffer[j] * alpha / 255 + cur_blue  * (255 - alpha) / 255;
         }
         j += 2;
         break;

      case 3:
         rgbbuffer[k++] = buffer[j++];
         rgbbuffer[k++] = buffer[j++];
         rgbbuffer[k++] = buffer[j++];
         break;

      case 4:
         if (pass > 0 && mode > 0) {
            src1[0] = cur_red;
            src1[1] = cur_green;
            src1[2] = cur_blue;
            src1[3] = 255;
            src2[0] = buffer[j];
            src2[1] = buffer[j + 1];
            src2[2] = buffer[j + 2];
            src2[3] = buffer[j + 3];
            pixel_apply_layer_mode(src1, src2, &buffer[j], 4, 4, TRUE, TRUE, mode);
         }
         alpha = buffer[j + 3];
         rgbbuffer[k++] = buffer[j++] * alpha / 255 + cur_red   * (255 - alpha) / 255;
         rgbbuffer[k++] = buffer[j++] * alpha / 255 + cur_green * (255 - alpha) / 255;
         rgbbuffer[k++] = buffer[j++] * alpha / 255 + cur_blue  * (255 - alpha) / 255;
         j++;
         break;

      default:
         break;
      }
   }

   return FALSE;
}


GimvImage *
gimv_image_rgba2rgb (GimvImage *image,
                     gint       bg_red,
                     gint       bg_green,
                     gint       bg_blue,
                     gboolean   ignore_alpha)
{
   GimvImage *dest_image = NULL;
   gint width, height, i, j;
   guchar *src, *dest;

   /* set bg color */
   if (bg_red >= 0 && bg_red < 256)
      bgcolor_red = bg_red;
   if (bg_green >= 0 && bg_green < 256)
      bgcolor_green = bg_green;
   if (bg_blue >= 0 && bg_blue < 256)
      bgcolor_blue = bg_blue;

#ifdef HAVE_GDK_IMLIB
   goto ERROR;
#endif /* HAVE_GDK_IMLIB */

   if (!gimv_image_has_alpha (image)) goto ERROR;

   width  = gimv_image_width  (image);
   height = gimv_image_height (image);
   src = gimv_image_get_pixels (image);

   dest = g_malloc (width * height * 3 * sizeof (guchar));

   for (i = 0; i < height; i++) {
      if (ignore_alpha) {
         for (j = 0; j < width; j++) {
            gint pos, src_pos;
            src_pos  = (j + i * width) * 4;
            pos = (j + i * width) * 3;
            dest [pos]     = src [src_pos];
            dest [pos + 1] = src [src_pos + 1];
            dest [pos + 2] = src [src_pos + 2];
         }
      } else {
        gimv_image_add_layer (src + (width * i * 4),
                               width, 0, 4,
                               0, NORMAL_MODE,
                               dest + (width * i * 3));
      }
   }

   dest_image = gimv_image_create_from_data (dest, width, height, FALSE);

ERROR:
   /* reset bg color to default value */
   bgcolor_red   = 255;
   bgcolor_green = 255;
   bgcolor_blue  = 255;

   return dest_image;
}
