/* -*- Mode: C; tab-width: 3; indent-tabs-mode: nil; c-basic-offset: 3 -*- */

/*
 * GImageView
 * Copyright (C) 2001-2002 Takuro Ashie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: image_loader.c,v 1.14.2.7 2003/01/29 21:27:37 makeinu Exp $
 */

#include "image_loader.h"
#include "fileutil.h"

/* loader */
#include "prefs.h"
#include "pixbuf_loader.h"
#include "imlib_loader.h"
#include "png_loader.h"
#include "bmp.h"
#include "pcx.h"
#include "xbm.h"
#include "mag.h"
#include "tga.h"
#include "xcf.h"
#include "xvpics.h"
#include "mng.h"
#include "mplayer_image_loader.h"
#include "svg.h"
#include "spi_image_loader.h"
#include "wmf.h"

#if defined (HAVE_GDK_PIXBUF)
#  include <gdk-pixbuf/gdk-pixbuf.h>
#  include "pixbuf_utils.h"
typedef GdkPixbuf GimvRawImage;
#elif defined (HAVE_GDK_IMLIB)
#  include <gdk_imlib.h>
typedef GdkImlibImage GimvRawImage;
#endif /* HAVE_GDK_PIXBUF */


typedef enum {
   LOAD_START_SIGNAL,
   PROGRESS_UPDATE_SIGNAL,
   LOAD_END_SIGNAL,
   LAST_SIGNAL
} ImageLoaderSignalType;


typedef GimvImage *(*ImageLoaderFunc)    (ImageLoader *loader,
                                          gpointer     data);


static void      image_loader_class_init (ImageLoaderClass *klass);
static void      image_loader_init       (ImageLoader      *loader);
static void      image_loader_destroy    (GtkObject        *object);
#ifdef ENABLE_PTHREAD
static void     *image_loader_thread     (void             *this);
#endif /* ENABLE_PTHREAD */
static gboolean  idle_image_loader_load  (gpointer          data);

/* callback */
static void      image_loader_load_end   (ImageLoader      *loader);


static ImageLoaderFunc loader_funcs[] =
{
   /* use gio interface */
   gimv_png_load,
   bmp_load,
   pcx_load,
   tga_load,
   mag_load,
   xvpics_load,
   xcf_load,

#ifdef ENABLE_MNG
   mng_load,
#endif /* ENABLE_MNG */

#if HAVE_GDK_PIXBUF
   pixbuf_load,
#endif /* HAVE_GDK_PIXBUF */

#ifdef ENABLE_WMF
   gimv_wmf_load,
#endif /* ENABLE_WMF */

   /* use filename interface */
   xbm_load,

#ifdef ENABLE_SVG
   svg_load_image,
#endif /* ENABLE_SVG */

#if HAVE_GDK_PIXBUF
   pixbuf_load_file,
#elif defined (HAVE_GDK_IMLIB)
   gimv_imlib_load_file,
#endif /* HAVE_GDK_PIXBUF */

#ifdef ENABLE_SPI
   spi_load_image,
#endif /* ENABLE_SPI */
};
static gint loader_funcs_num
= sizeof (loader_funcs) / sizeof (ImageLoaderFunc);


static GtkObjectClass *parent_class = NULL;
static gint image_loader_signals[LAST_SIGNAL] = {0};


static void
image_loader_class_init (ImageLoaderClass *klass)
{
   GtkObjectClass *object_class;

   object_class = (GtkObjectClass *) klass;
   parent_class = gtk_type_class (gtk_object_get_type ());

   image_loader_signals[LOAD_START_SIGNAL]
      = gtk_signal_new ("load_start",
                        GTK_RUN_FIRST,
                        GTK_CLASS_TYPE (object_class),
                        GTK_SIGNAL_OFFSET (ImageLoaderClass, load_start),
                        gtk_signal_default_marshaller,
                        GTK_TYPE_NONE, 0);

   image_loader_signals[PROGRESS_UPDATE_SIGNAL]
      = gtk_signal_new ("progress_update",
                        GTK_RUN_FIRST,
                        GTK_CLASS_TYPE (object_class),
                        GTK_SIGNAL_OFFSET (ImageLoaderClass, load_end),
                        gtk_signal_default_marshaller,
                        GTK_TYPE_NONE, 0);

   image_loader_signals[LOAD_END_SIGNAL]
      = gtk_signal_new ("load_end",
                        GTK_RUN_FIRST,
                        GTK_CLASS_TYPE (object_class),
                        GTK_SIGNAL_OFFSET (ImageLoaderClass, load_end),
                        gtk_signal_default_marshaller,
                        GTK_TYPE_NONE, 0);

   gtk_object_class_add_signals (object_class,
                                 image_loader_signals, LAST_SIGNAL);

   object_class->destroy  = image_loader_destroy;

   klass->load_start      = NULL;
   klass->progress_update = NULL;
   klass->load_end        = image_loader_load_end;
}


static void
image_loader_init (ImageLoader *loader)
{
   loader->info              = NULL;
   loader->gio               = NULL;
   loader->image             = NULL;
   loader->load_as_animation = FALSE;
   loader->width_scale       = 1.0;
   loader->height_scale      = 1.0;
   loader->temp_file         = NULL;
   loader->flags             = 0;
   loader->next_info         = NULL;

#ifdef USE_GTK2
   gtk_object_ref (GTK_OBJECT (loader));
   gtk_object_sink (GTK_OBJECT (loader));
#endif
}


guint
image_loader_get_type (void)
{
   static guint image_loader_type = 0;

   if (!image_loader_type) {
      static const GtkTypeInfo image_loader_info = {
         "ImageLoader",
         sizeof (ImageLoader),
         sizeof (ImageLoaderClass),
         (GtkClassInitFunc) image_loader_class_init,
         (GtkObjectInitFunc) image_loader_init,
         NULL,
         NULL,
         (GtkClassInitFunc) NULL,
      };

      image_loader_type = gtk_type_unique (gtk_object_get_type (), &image_loader_info);
   }

   return image_loader_type;
}


ImageLoader *
image_loader_new (void)
{
   ImageLoader *loader = IMAGE_LOADER (gtk_type_new (image_loader_get_type ()));

   return loader;
}


ImageLoader *
image_loader_new_with_image_info (ImageInfo *info)
{
   ImageLoader *loader;

   loader = image_loader_new ();
   image_loader_set_image_info (loader, info);

   return loader;
}


ImageLoader *
image_loader_new_with_file_name (const gchar *filename)
{
   ImageLoader *loader;
   ImageInfo *info;

   info = image_info_get (filename);
   if (!info) return NULL;
   loader = image_loader_new_with_image_info (info);

   return loader;
}


ImageLoader *
image_loader_ref (ImageLoader *loader)
{
   g_return_val_if_fail (IS_IMAGE_LOADER (loader), NULL);

   gtk_object_ref (GTK_OBJECT (loader));

   return loader;
}


void
image_loader_unref (ImageLoader *loader)
{
   g_return_if_fail (IS_IMAGE_LOADER (loader));

   gtk_object_unref (GTK_OBJECT (loader));
}


static void
image_loader_destroy (GtkObject *object)
{
   ImageLoader *loader = IMAGE_LOADER (object);

   if (image_loader_is_loading (loader))
      image_loader_load_stop (loader);

   if (loader->info)
      image_info_unref (loader->info);
   loader->info = NULL;

   if (loader->next_info)
      image_info_unref (loader->next_info);
   loader->next_info = NULL;

   if (loader->gio)
      gimv_io_unref (loader->gio);
   loader->gio = NULL;

   if (loader->image)
      gimv_image_unref (loader->image);
   loader->image = NULL;

   if (loader->temp_file)
      g_free ((gchar *) loader->temp_file);
   loader->temp_file = NULL;

   if (GTK_OBJECT_CLASS (parent_class)->destroy)
      (*GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}


void
image_loader_set_image_info (ImageLoader *loader, ImageInfo *info)
{
   g_return_if_fail (IS_IMAGE_LOADER (loader));
   g_return_if_fail (info);

   if (image_loader_is_loading (loader)) {
      image_loader_load_stop (loader);
      if (loader->next_info)
         image_info_unref (loader->next_info);
      loader->next_info = image_info_ref (info);
      return;
   }

   if (loader->image)
      gimv_image_unref (loader->image);
   loader->image = NULL;

   if (loader->temp_file)
      g_free ((gchar *)loader->temp_file);
   loader->temp_file = NULL;

   if (loader->info)
      image_info_unref (loader->info);

   if (info)
      loader->info = image_info_ref (info);
   else
      loader->info = NULL;
}


gboolean
image_loader_set_as_animation (ImageLoader *loader,
                               gboolean animation)
{
   g_return_val_if_fail (IS_IMAGE_LOADER (loader), FALSE);
   g_return_val_if_fail (!image_loader_is_loading (loader), FALSE);

   loader->load_as_animation = animation;

   return TRUE;
}


gboolean
image_loader_set_scale (ImageLoader *loader,
                        gfloat       w_scale,
                        gfloat       h_scale)
{
   g_return_val_if_fail (IS_IMAGE_LOADER (loader), FALSE);
   g_return_val_if_fail (!image_loader_is_loading (loader), FALSE);

   loader->width_scale  = w_scale;
   loader->height_scale = h_scale;

   return TRUE;
}


void
image_loader_load_start (ImageLoader *loader)
{
   g_return_if_fail (IS_IMAGE_LOADER (loader));
   g_return_if_fail (loader->info);

   g_return_if_fail (!image_loader_is_loading(loader));

   loader->flags &= ~IMAGE_LOADER_LOADING_FLAG;
   loader->flags &= ~IMAGE_LOADER_CANCEL_FLAG;

   /* gtk_idle_add (idle_image_loader_load, loader); */
   image_loader_load (loader);
}


void
image_loader_load_stop (ImageLoader *loader)
{
   g_return_if_fail (IS_IMAGE_LOADER (loader));

   if (loader->flags & IMAGE_LOADER_LOADING_FLAG) {
      loader->flags |= IMAGE_LOADER_CANCEL_FLAG;
   }
}


gboolean
image_loader_is_loading (ImageLoader *loader)
{
   g_return_val_if_fail (IS_IMAGE_LOADER (loader), FALSE);

   if (loader->flags & IMAGE_LOADER_LOADING_FLAG)
      return TRUE;
   else
      return FALSE;
}


GimvImage *
image_loader_get_image (ImageLoader *loader)
{
   g_return_val_if_fail (IS_IMAGE_LOADER (loader), NULL);

   return loader->image;
}


void
image_loader_unref_image (ImageLoader *loader)
{
   g_return_if_fail (IS_IMAGE_LOADER (loader));

   if (loader->image)
      gimv_image_unref (loader->image);
   loader->image = NULL;
}


GimvIO *
image_loader_get_gio (ImageLoader *loader)
{
   g_return_val_if_fail (IS_IMAGE_LOADER (loader), NULL);

   return loader->gio;
}


const gchar *
image_loader_get_path (ImageLoader *loader)
{
   g_return_val_if_fail (IS_IMAGE_LOADER (loader), NULL);

   if (image_info_need_temp_file (loader->info)) {
      if (!loader->temp_file)
         loader->temp_file = image_info_get_temp_file_path (loader->info);
      return loader->temp_file;
   }

   return image_info_get_path (loader->info);
}


gboolean
image_loader_load_as_animation (ImageLoader *loader)
{
   g_return_val_if_fail (IS_IMAGE_LOADER (loader), FALSE);

   return loader->load_as_animation;
}


void
image_loader_get_scale (ImageLoader *loader,
                        gfloat *width_scale_ret,
                        gfloat *height_scale_ret)
{
   g_return_if_fail (IS_IMAGE_LOADER (loader));
   g_return_if_fail (width_scale_ret && height_scale_ret);

   *width_scale_ret  = loader->width_scale;
   *height_scale_ret = loader->height_scale;
}


gboolean
image_loader_progress_update (ImageLoader *loader)
{
   g_return_val_if_fail (IS_IMAGE_LOADER (loader), FALSE);

   gtk_signal_emit (GTK_OBJECT(loader),
                    image_loader_signals[PROGRESS_UPDATE_SIGNAL]);

   if (loader->flags & IMAGE_LOADER_CANCEL_FLAG)
      return FALSE;
   else
      return TRUE;
}


void
image_loader_load (ImageLoader *loader)
{
   const gchar *filename;
   gint i;
   gboolean cancel = FALSE;

   g_return_if_fail (IS_IMAGE_LOADER (loader));

   if (image_loader_is_loading (loader)) {
      return;
   }

   loader->flags &= ~IMAGE_LOADER_LOADING_FLAG;
   loader->flags &= ~IMAGE_LOADER_CANCEL_FLAG;
   loader->flags |= IMAGE_LOADER_LOADING_FLAG;

   /* load the image */
   if (image_info_need_temp_file (loader->info)) {
      g_free ((gchar *) loader->temp_file);
      loader->temp_file = image_info_get_temp_file (loader->info);
      g_return_if_fail (loader->temp_file);
   }

   filename = image_loader_get_path (loader);

   if (!file_exists (filename)) return;

   if (loader->image)
      gimv_image_unref (loader->image);
   loader->image = NULL;

   gtk_signal_emit (GTK_OBJECT(loader),
                    image_loader_signals[LOAD_START_SIGNAL]);

#ifdef ENABLE_MPLAYER
   /* FIXME!! ad-hoc */
   if (image_info_is_movie (loader->info)) {
      if (conf.mplayer_thumbnail_enable)
         loader->image = gimv_mplayer_load_file (loader, NULL);
   } else
#endif /* ENABLE_MPLAYER */
   {
      /* for GimvIO interface */
      loader->gio = gimv_io_new (filename, "rb");

      for (i = 0; i < loader_funcs_num; i++) {
         cancel = !image_loader_progress_update (loader);
         if (cancel) break;

         gimv_io_seek (loader->gio, 0, SEEK_SET);

         loader->image = loader_funcs[i] (loader, NULL);

         if (loader->image) break;
         if (loader->flags & IMAGE_LOADER_CANCEL_FLAG) break;
      }

      if (loader->gio)
         gimv_io_close (loader->gio);
      loader->gio = NULL;
   }

   loader->flags &= ~IMAGE_LOADER_LOADING_FLAG;
   if (loader->flags & IMAGE_LOADER_CANCEL_FLAG) {
      loader->flags &= ~IMAGE_LOADER_CANCEL_FLAG;
      if (loader->flags & IMAGE_LOADER_DEBUG_FLAG)
         g_print ("----- loading canceled -----\n");
      /* emit canceled signal? */
      gtk_signal_emit (GTK_OBJECT(loader),
                       image_loader_signals[LOAD_END_SIGNAL]);
   } else {
      if (loader->flags & IMAGE_LOADER_DEBUG_FLAG)
         g_print ("----- loading done -----\n");
      gtk_signal_emit (GTK_OBJECT(loader),
                       image_loader_signals[LOAD_END_SIGNAL]);
   }
}


#ifdef ENABLE_PTHREAD
static void *
image_loader_thread (void *this)
{
   ImageLoader *loader = this;

   image_loader_load (loader);

   return NULL;
}
#endif /* ENABLE_PTHREAD */


static gboolean
idle_image_loader_load (gpointer data)
{
   ImageLoader *loader = data;

   image_loader_load (loader);

   return FALSE;
}


static void
image_loader_load_end (ImageLoader *loader)
{
   if (!loader->next_info)
      return;

   loader->info = loader->next_info;
   loader->next_info = NULL;
   gtk_idle_add (idle_image_loader_load, loader);
}
