/* -*- Mode: C; tab-width: 3; indent-tabs-mode: nil; c-basic-offset: 3 -*- */

/*
 * GImageView
 * Copyright (C) 2001-2002 Takuro Ashie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: imgwin_fullscreen.c,v 1.6.2.2 2003/01/09 16:19:41 makeinu Exp $
 */

#include "imgwin_fullscreen.h"
#include "cursors.h"
#include "menu.h"
#include "prefs.h"


static void
show_cursor (ImageWindow *iw)
{
   GdkCursor *cursor;

   g_return_if_fail (iw);
   g_return_if_fail (iw->iv);
   g_return_if_fail (iw->iv->draw_area);
   g_return_if_fail (GTK_WIDGET_MAPPED (iw->iv->draw_area));

   cursor = cursor_get (iw->iv->draw_area->window,
                        CURSOR_HAND_OPEN);
   gdk_window_set_cursor (iw->iv->draw_area->window, 
                          cursor);
   gdk_cursor_destroy (cursor);
}


static void
hide_cursor (ImageWindow *iw)
{
   GdkCursor *cursor;

   g_return_if_fail (iw);
   g_return_if_fail (iw->iv);
   g_return_if_fail (iw->iv->draw_area);
   g_return_if_fail (GTK_WIDGET_MAPPED (iw->iv->draw_area));

   cursor = cursor_get (iw->iv->draw_area->window,
                        CURSOR_VOID);
   gdk_window_set_cursor (iw->iv->draw_area->window, 
                          cursor);
   gdk_cursor_destroy (cursor);
}


static gint 
timeout_hide_cursor (gpointer data)
{
   ImageWindow *iw = data;

   hide_cursor (iw);
   iw->hide_cursor_timer_id = 0;

   return FALSE;
}


static gboolean
cb_fullscreen_key_press (GtkWidget *widget, 
                         GdkEventKey *event,
                         ImageWindow *iw)
{
   g_return_val_if_fail (iw, FALSE);
        
   switch (event->keyval) {
   case GDK_Escape:
      if (iw->flags & ImgWinCreatingFlag) return TRUE;
      menu_check_item_set_active (iw->view_menu, "/Full Screen", FALSE);
      return TRUE;
   default:
      break;
   }

#ifdef USE_GTK2
   gtk_accel_groups_activate (G_OBJECT (iw->window),
                              event->keyval, event->state);
#else /* USE_GTK2 */
   gtk_accel_groups_activate (GTK_OBJECT (iw->window),
                              event->keyval, event->state);
#endif /* USE_GTK2 */

   return TRUE;
}


static gboolean
cb_fullscreen_motion_notify (GtkWidget *widget, 
                             GdkEventButton *bevent, 
                             gpointer data)
{
   ImageWindow *iw = data;

   gdk_window_get_pointer (widget->window, NULL, NULL, NULL);
   show_cursor (iw);

   if (iw->hide_cursor_timer_id != 0)
      gtk_timeout_remove (iw->hide_cursor_timer_id);
   iw->hide_cursor_timer_id
      = gtk_timeout_add (IMGWIN_FULLSCREEN_HIDE_CURSOR_DELAY,
                         timeout_hide_cursor,
                         iw);

   return FALSE;
}


static void
set_menu_sensitive (ImageWindow *iw, gboolean sensitive)
{
   g_return_if_fail (iw);

   menu_item_set_sensitive (iw->menubar,   "/File/Open...",     sensitive);
   menu_item_set_sensitive (iw->menubar,   "/File/Open Image Window",     sensitive);
   menu_item_set_sensitive (iw->menubar,   "/File/Open Thumbnail Window", sensitive);
   menu_item_set_sensitive (iw->menubar,   "/File/Quit",        sensitive);
   menu_item_set_sensitive (iw->menubar,   "/Edit/Options...",  sensitive);
   menu_item_set_sensitive (iw->view_menu, "/Menu Bar",         sensitive);
   menu_item_set_sensitive (iw->view_menu, "/Tool Bar",         sensitive);
   menu_item_set_sensitive (iw->view_menu, "/Slide Show Player",sensitive);
   menu_item_set_sensitive (iw->view_menu, "/Status Bar",       sensitive);
   menu_item_set_sensitive (iw->view_menu, "/Scroll Bar",       sensitive);
   menu_item_set_sensitive (iw->view_menu, "/Fit to Image",     sensitive);
   menu_item_set_sensitive (iw->view_menu, "/Maximize",         sensitive);
   gtk_widget_set_sensitive (iw->help_menu, sensitive);
}


void
imgwin_fullscreen_new (ImageWindow *iw)
{ 
   g_return_if_fail (iw);

   iw->flags |= ImgWinFullScreenFlag;

   /* create window */
   iw->fullscreen = gtk_window_new (GTK_WINDOW_POPUP);

   gtk_window_set_wmclass (GTK_WINDOW (iw->fullscreen), "",
                           "gimv_fullscreen");
   gtk_widget_set_uposition (iw->fullscreen, 0, 0);
   gtk_widget_set_usize (iw->fullscreen, 
                         gdk_screen_width (), 
                         gdk_screen_height ());

   gtk_signal_connect (GTK_OBJECT (iw->fullscreen), 
                       "key_press_event",
                       (GtkSignalFunc) cb_fullscreen_key_press, 
                       iw);
   gtk_signal_connect (GTK_OBJECT (iw->fullscreen),
                       "motion_notify_event",
                       (GtkSignalFunc) cb_fullscreen_motion_notify,
                       iw);

   /* set draw widget */
   if (!(iw->slideshow && conf.slideshow_set_bg)
       && conf.imgwin_fullscreen_set_bg)
   {
      GtkStyle *style;
      GdkColor *color = g_new0 (GdkColor, 1);

      /* save current color */
      style = gtk_widget_get_style (iw->iv->draw_area);
      *color = style->bg[GTK_STATE_NORMAL];
      gtk_object_set_data_full (GTK_OBJECT (iw->fullscreen),
                                "ImgWinFullScreen::OrigColor",
                                color,
                                (GtkDestroyNotify) g_free);

      /* set bg color */
      imageview_set_bg_color (iw->iv,
                              conf.imgwin_fullscreen_bg_color[0],
                              conf.imgwin_fullscreen_bg_color[1],
                              conf.imgwin_fullscreen_bg_color[2]);
   }

   gtk_widget_show (iw->fullscreen);

   imageview_set_fullscreen (iw->iv, GTK_WINDOW (iw->fullscreen));

   set_menu_sensitive (iw, FALSE);

   gdk_keyboard_grab (iw->fullscreen->window, TRUE, GDK_CURRENT_TIME);
   gtk_grab_add (iw->fullscreen);
   gtk_widget_grab_focus (GTK_WIDGET (iw->iv));

   /* enable auto hide cursor stuff */
   iw->hide_cursor_timer_id
      = gtk_timeout_add (IMGWIN_FULLSCREEN_HIDE_CURSOR_DELAY,
                         timeout_hide_cursor,
                         iw);
}


void
imgwin_fullscreen_delete (ImageWindow *iw)
{
   GdkColor *color;

   g_return_if_fail (iw);

   /* restore draw widget */
   color = gtk_object_get_data (GTK_OBJECT (iw->fullscreen),
                                "ImgWinFullScreen::OrigColor");
   if (color)
      imageview_set_bg_color (iw->iv, color->red, color->green, color->blue);

   imageview_unset_fullscreen (iw->iv);

   gtk_widget_grab_focus (GTK_WIDGET (iw->iv));

   /* disable auto hide cursor stuff */
   if (iw->hide_cursor_timer_id != 0)
      gtk_timeout_remove (iw->hide_cursor_timer_id);
   show_cursor (iw);

   /* restore sensitivity */
   set_menu_sensitive (iw, TRUE);

   gtk_widget_destroy (iw->fullscreen);
   iw->fullscreen = NULL;
   iw->flags &= ~ImgWinFullScreenFlag;

   if (iw->slideshow)
      gtk_widget_destroy (iw->window);
}
