/* -*- Mode: C; tab-width: 3; indent-tabs-mode: nil; c-basic-offset: 3 -*- */

/*
 * GImageView
 * Copyright (C) 2001 Takuro Ashie
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: plugin.c,v 1.23 2002/10/17 22:50:04 makeinu Exp $
 */

#include <stdio.h>
#include <string.h>
#include <gmodule.h>

#include "gimageview.h"

#include "fileutil.h"
#include "gfileutil.h"
#include "plugin.h"
#include "prefs.h"

#include "fr-archive.h"
#include "image_saver.h"
#include "thumbnail_support.h"

#define PLUGIN_RC_DIR       "pluginrc"
#define ARCHIVER_RC_FILE    "archiver"
#define IMAGE_SAVER_RC_FILE "image_saver"
#define THUMB_CACHE_RC_FILE "thumbnail"


static GList *search_dir_list     = NULL;

static GHashTable *plugins        = NULL;
static GList *archiver_plugins    = NULL;
static GList *image_saver_plugins = NULL;
static GList *thumb_cache_plugins = NULL;

GHashTable *archiver_plugin_prefs = NULL;
GHashTable *image_saver_plugin_prefs = NULL;
GHashTable *thumb_cache_plugin_prefs = NULL;


/******************************************************************************
 *
 *   Private functions.
 *
 ******************************************************************************/
static void
plugin_clear_search_dir_list (void)
{
   g_list_foreach (search_dir_list, (GFunc) g_free, NULL);
   g_list_free (search_dir_list);
   search_dir_list = NULL;
}


static gchar *
get_plugin_prefs_filename (PluginType type)
{
   switch (type) {
   case PLUGIN_EXT_ARCHIVER:
      return ARCHIVER_RC_FILE;
   case PLUGIN_IMAGE_SAVER:
      return IMAGE_SAVER_RC_FILE;
   case PLUGIN_THUMB_CACHE:
      return THUMB_CACHE_RC_FILE;
   default:
      break;
   }

   return NULL;
}


static gchar *
get_plugin_prefs_path (PluginType type)
{
   gchar *filename, *path;

   filename = get_plugin_prefs_filename (type);
   g_return_val_if_fail (filename, NULL);

   path = g_strconcat (g_getenv("HOME"), "/", RC_DIR, "/",
                       PLUGIN_RC_DIR, "/", filename, NULL);

   return path;
}


static GHashTable *
get_prefs_table (PluginType type)
{
   switch (type) {
   case PLUGIN_EXT_ARCHIVER:
      if (!archiver_plugin_prefs)
         archiver_plugin_prefs = g_hash_table_new (g_str_hash, g_str_equal);
      return archiver_plugin_prefs;

   case PLUGIN_IMAGE_SAVER:
      if (!image_saver_plugin_prefs)
         image_saver_plugin_prefs = g_hash_table_new (g_str_hash, g_str_equal);
      return image_saver_plugin_prefs;

   case PLUGIN_THUMB_CACHE:
      if (!thumb_cache_plugin_prefs)
         thumb_cache_plugin_prefs = g_hash_table_new (g_str_hash, g_str_equal);
      return thumb_cache_plugin_prefs;

   default:
      break;
   }

   return NULL;
}


static void
plugin_prefs_read_file (PluginType type)
{
   FILE *pluginrc;
   gchar *path;
   gchar buf[BUF_SIZE], section[256], **pair = NULL;
   gint len;

   path = get_plugin_prefs_path (type);
   if (!path || !*path) return;

   pluginrc = fopen(path, "r");
   if (!pluginrc) {
      /* g_warning (_("Can't open plugin preference file for read.")); */
      goto ERROR;
   }

   section[0] = '\0';
   while (fgets (buf, sizeof(buf), pluginrc)) {
      g_strstrip (buf);
      if (buf[0] == '\0') break;
      if (buf[0] == '\n') continue;

      len = strlen(buf);

      if (len > 4 && len < 256 && buf[0] == '[' && buf[len - 1] == ']') {
         strncpy (section, buf + 1, len - 2);
         section[len - 2] = '\0';
         continue;
      }

      if (!*section) continue;

      pair = g_strsplit (buf, "=", 2);
      if (!pair) continue;

      if (pair[0]) g_strstrip(pair[0]);
      if (pair[1]) g_strstrip(pair[1]);

      if (pair[0] && *pair[0]) {
         plugin_prefs_save_value (section, type, pair[0], pair[1]);
      }

      g_strfreev (pair);
      pair = NULL;
   }

   fclose (pluginrc);
ERROR:
   g_free (path);
}


static void
write_value (gpointer key, gpointer value, gpointer user_data)
{
   FILE *pluginrc = user_data;
   gchar *key_string = key;
   gchar *value_string = value;

   g_return_if_fail (key_string && *key_string);
   g_return_if_fail (value_string);
   g_return_if_fail (pluginrc);

   fprintf (pluginrc, "%s=%s\n", key_string, value_string);
}


static void
write_section (gpointer key, gpointer value, gpointer user_data)
{
   GHashTable *htable = value;
   FILE *pluginrc = user_data;
   gchar *string = key;

   g_return_if_fail (string && *string);
   g_return_if_fail (pluginrc);

   if (!htable) return;
   if (g_hash_table_size (htable) < 1) return;

   fprintf (pluginrc, "[%s]\n", string);
   g_hash_table_foreach (htable, (GHFunc) write_value, pluginrc);
   fprintf (pluginrc, "\n");
}


static void
plugin_prefs_write_file (PluginType type)
{
   GHashTable *prefs_table;
   gchar *path;
   FILE *pluginrc;

   prefs_table = get_prefs_table (type);
   if (!prefs_table) return;
   if (g_hash_table_size (prefs_table) < 1) return;

   path = get_plugin_prefs_path (type);
   if (!path) return;

   mkdirs (path);

   pluginrc = fopen (path, "w");
   if (!pluginrc) {
      g_warning (_("Can't open plugin preference file for write."));
      goto ERROR;
   }

   g_hash_table_foreach (prefs_table, (GHFunc) write_section, pluginrc);

   fclose (pluginrc);
ERROR:
   g_free (path);
}


/******************************************************************************
 *
 *   Public functions.
 *
 ******************************************************************************/
void
plugin_search_directory (const gchar *dirname)
{
   GModule *module;
   gboolean success, regist = FALSE;
   PluginInfo *info;
   GList *filelist = NULL, *node;

   g_return_if_fail (dirname);
   if (!isdir (dirname)) return;

   get_dir (dirname, GETDIR_FOLLOW_SYMLINK, &filelist, NULL);
   if (!filelist) return;

   for (node = filelist; node; node = g_list_next (node)) {
      gchar *filename = node->data;
      gint len;

      success = FALSE;
      regist = FALSE;

      len = strlen (filename);
      if (len < 3 || strcmp (filename + len - 3, ".so")) continue;

      /* load module */
      module =  g_module_open (filename, G_MODULE_BIND_LAZY);
      success = g_module_symbol (module, "gimv_plugin_info", (gpointer) &info);

      if (!success) continue;

      /* regist each method to memory */
      switch (info->type) {
      case PLUGIN_EXT_ARCHIVER:
         regist = fr_archive_regist_archiver (info->name, g_module_name (module),
                                              info->n_entries, info->implement);
         if (regist)
            archiver_plugins = g_list_append (archiver_plugins, module);
         break;

      case PLUGIN_IMAGE_SAVER:
         regist = image_saver_regist (info->name, g_module_name (module),
                                      info->n_entries, info->implement);
         if (regist)
            image_saver_plugins = g_list_append (image_saver_plugins, module);
         break;

      case PLUGIN_THUMB_CACHE:
         regist = thumbsupport_regist_loader (info->name, g_module_name (module),
                                              info->n_entries, info->implement);
         if (regist)
            thumb_cache_plugins = g_list_append (thumb_cache_plugins, module);
         break;

      default:
         break;
      }

      if (!regist)
         g_module_close (module);
      else
         g_hash_table_insert (plugins, (gchar *) g_module_name (module), module);
   }

   g_list_foreach (filelist, (GFunc) g_free, NULL);
   g_list_free (filelist);
}


void
plugin_create_search_dir_list (void)
{
   gchar **dirs;
   gint i;

   plugin_clear_search_dir_list ();

   if (!conf.plugin_search_dir_list || !*conf.plugin_search_dir_list) return;

   dirs = g_strsplit (conf.plugin_search_dir_list, ",", -1);

   for (i = 0; dirs && dirs[i]; i++) {
      if (*dirs[i])
         search_dir_list = g_list_append (search_dir_list, g_strdup (dirs[i]));
   }

   g_strfreev (dirs);
}


GList *
plugin_get_search_dir_list (void)
{
   return search_dir_list;
}


const gchar *
plugin_get_name (GModule *module)
{
   gboolean success;
   PluginInfo *info;

   g_return_val_if_fail (module, NULL);

   success = g_module_symbol (module, "gimv_plugin_info", (gpointer) &info);
   g_return_val_if_fail (success, NULL);

   return info->name;
}


const gchar *
plugin_get_version_string (GModule *module)
{
   gboolean success;
   PluginInfo *info;

   g_return_val_if_fail (module, NULL);

   success = g_module_symbol (module, "gimv_plugin_info", (gpointer) &info);
   g_return_val_if_fail (success, NULL);

   return info->version;
}


const gchar *
plugin_get_author (GModule *module)
{
   gboolean success;
   PluginInfo *info;

   g_return_val_if_fail (module, NULL);

   success = g_module_symbol (module, "gimv_plugin_info", (gpointer) &info);
   g_return_val_if_fail (success, NULL);

   return info->author;
}


const gchar *
plugin_get_module_name (GModule *module)
{
   gboolean success;
   PluginInfo *info;

   g_return_val_if_fail (module, NULL);

   success = g_module_symbol (module, "gimv_plugin_info", (gpointer) &info);
   g_return_val_if_fail (success, NULL);

   return g_module_name (module);
}


GList *
plugin_get_list (PluginType type)
{
   switch (type) {
   case PLUGIN_EXT_ARCHIVER:
      return archiver_plugins;
   case PLUGIN_IMAGE_SAVER:
      return image_saver_plugins;
   case PLUGIN_THUMB_CACHE:
      return thumb_cache_plugins;
   default:
      break;
   }

   return NULL;
}


gboolean
plugin_prefs_save_value (const gchar *pname,
                         PluginType   ptype,
                         const gchar *key,
                         const gchar *value)
{
   GHashTable *prefs_table, *htable;
   gchar *old_value, *orig_key;
   gboolean success;

   g_return_val_if_fail (pname && *pname, FALSE);
   g_return_val_if_fail (key && *key, FALSE);
   g_return_val_if_fail (value, FALSE);

   prefs_table = get_prefs_table (ptype);
   g_return_val_if_fail (prefs_table, FALSE);

   htable = g_hash_table_lookup (prefs_table, pname);
   if (!htable) {
      htable = g_hash_table_new (g_str_hash, g_str_equal);
      g_hash_table_insert (prefs_table, g_strdup (pname), htable);
   }

   success = g_hash_table_lookup_extended (htable, key,
                                           (gpointer) &orig_key,
                                           (gpointer) &old_value);
   if (success) {
      g_hash_table_remove (htable, key);
      g_free (orig_key);
      g_free (old_value);
   }

   g_hash_table_insert (htable, g_strdup (key), g_strdup (value));

   return TRUE;
}


gboolean
plugin_prefs_load_value (const gchar *pname,
                         PluginType   ptype,
                         const gchar *key,
                         PluginPrefsType type,
                         gpointer data)
{
   GHashTable *prefs_table, *htable;
   gchar *value;

   g_return_val_if_fail (pname && *pname, FALSE);
   g_return_val_if_fail (key && *key, FALSE);
   g_return_val_if_fail (data, FALSE);

   prefs_table = get_prefs_table (ptype);
   g_return_val_if_fail (prefs_table, FALSE);

   htable = g_hash_table_lookup (prefs_table, pname);
   if (!htable) return FALSE;

   value = g_hash_table_lookup (htable, key);
   if (!value) return FALSE;
   /* g_return_val_if_fail (!*value, FALSE); */

   switch (type) {
   case PLUGIN_PREFS_STRING:
      *((gchar **) data) = value;
      break;
   case PLUGIN_PREFS_INT:
      *((gint *) data) = atoi (value);
      break;
   case PLUGIN_PREFS_FLOAT:
      *((gfloat *) data) = atof (value);
      break;
   case PLUGIN_PREFS_BOOL:
      if (!g_strcasecmp (value, "TRUE"))
         *((gboolean *) data) = TRUE;
      else
         *((gboolean *) data) = FALSE;
      break;
   default:
      return FALSE;
   }

   return TRUE;
}


void
plugin_prefs_read_files (void)
{
   plugin_prefs_read_file (PLUGIN_EXT_ARCHIVER);
   plugin_prefs_read_file (PLUGIN_IMAGE_SAVER);
   plugin_prefs_read_file (PLUGIN_THUMB_CACHE);
}


void
plugin_prefs_write_files (void)
{
   plugin_prefs_write_file (PLUGIN_EXT_ARCHIVER);
   plugin_prefs_write_file (PLUGIN_IMAGE_SAVER);
   plugin_prefs_write_file (PLUGIN_THUMB_CACHE);
}


void
plugin_init (void)
{
   GList *list;

   plugin_create_search_dir_list ();

   /* check whether platform supports dynamic loading or not */
   if (!g_module_supported ()) return;

   if (!plugins)
      plugins = g_hash_table_new (g_str_hash, g_str_equal);

   /* load module from disk */
   list = g_list_last (search_dir_list);
   for (; list; list = g_list_previous (list)) {
      gchar *dir = list->data;

      if (dir && *dir)
         plugin_search_directory (dir);
   }
}
